#ifdef _WIN32
#include "stdafx.h"
#endif
#include <stdarg.h>		// Required for variable length arguments
#include <stdio.h>		// Required for file IO and NULL declaration
#include <stdlib.h>		// Required for memory manipulation
#include <time.h>
#include <BipCci.h>		// Common functions (between nodes and parsers)
#include <BipCni.h>		// Message processing nodes
#include "NodeUtils.h"

#ifdef _WIN32
#define LOGFILE "C:\\WKSPACE.TXT"
#else
#define LOGFILE "/tmp/WKSPACE.txt"
#endif

///////////////////////////////////////////////////////////
// mqsixCopyString
// Copy the CciChar strings ensuring there is enough storage
int mqsixStringCopy(CciChar *source, CciChar *destination, unsigned int iLength)
{
	// Ensure that there is enough space to copy the strings
	if (mqsixStringLengthBytes(source) > iLength)
	{
		return(1);
	}
	wcscpy(destination, source);
	return(0);
}

///////////////////////////////////////////////////////////
// mqsixStringLengthCharacters
// Return the number of characters contained within the
// string.  Does not include any terminator character ('\0')
unsigned int mqsixStringLengthCharacters(CciChar *string)
{
	if (string == NULL)
	{
		return(0);
	}
	return(wcslen(string));
}

///////////////////////////////////////////////////////////
// mqsixStringLengthBytes
// Return the length in bytes of memory that the string
// occupies.  This includes the trailing \0 character
// terminator.
unsigned int mqsixStringLengthBytes(CciChar *string)
{
	if (string == NULL)
	{
		return(0);
	}
	return(sizeof(CciChar) * (wcslen(string) + 1));
}

///////////////////////////////////////////////////////////
// mqsixStringCompare
int mqsixStringCompare(CciChar *stringOne, CciChar *stringTwo)
{
	return(wcscmp(stringOne, stringTwo));
}

///////////////////////////////////////////////////////////
// mqsixStringDuplicate
CciChar *mqsixStringDuplicate(CciChar *string)
{
	CciChar *temp;
	int		iSize;

	iSize = mqsixStringLengthBytes(string);
	temp = (CciChar *)malloc(iSize);
	mqsixStringCopy(string, temp, iSize);
	return(temp);
}

// **************
// * ATTRIBUTES *
// **************

///////////////////////////////////////////////////////////
// mqsixInitializeAttributes
void mqsixInitializeAttributes(MQSIX_ATTRIBUTES *pAttributes)
{
	int	i;
	pAttributes->iAttributeCount = 5;
	pAttributes->attributeNames = (CciChar **)malloc(sizeof(CciChar *) * pAttributes->iAttributeCount);
	pAttributes->attributeValues = (CciChar **)malloc(sizeof(CciChar *)* pAttributes->iAttributeCount);

	for (i=0; i<pAttributes->iAttributeCount; i++)
	{
		pAttributes->attributeNames[i] = NULL;
		pAttributes->attributeValues[i] = NULL;
	}

	// MQSI required attributes
	pAttributes->attributeNames[0] = mqsixStringDuplicate(L"label");
	pAttributes->attributeNames[1] = mqsixStringDuplicate(L"userTraceLevel");
	pAttributes->attributeValues[1] = mqsixStringDuplicate(L"none");
	pAttributes->attributeNames[2] = mqsixStringDuplicate(L"traceLevel");
	pAttributes->attributeValues[2] = mqsixStringDuplicate(L"none");
	pAttributes->attributeNames[3] = mqsixStringDuplicate(L"userTraceFilter");
	pAttributes->attributeValues[3] = mqsixStringDuplicate(L"none");
	pAttributes->attributeNames[4] = mqsixStringDuplicate(L"traceFilter");
	pAttributes->attributeValues[4] = mqsixStringDuplicate(L"none");
}

///////////////////////////////////////////////////////////
// mqsixReleaseAttributes
void mqsixReleaseAttributes(MQSIX_ATTRIBUTES *pAttributes)
{
	int	i;
	for (i=0; i<pAttributes->iAttributeCount; i++)
	{
		if (pAttributes->attributeValues[i] != NULL)
		{
			free(pAttributes->attributeValues[i]);
		}
	}
	free(pAttributes->attributeNames);
	free(pAttributes->attributeValues);
}

///////////////////////////////////////////////////////////
// mqsixGetAttributeName
int mqsixGetAttributeName(
	MQSIX_ATTRIBUTES	*pAttributes,
	int					index,
	CciChar				*buffer,
	int					bufsize)
{
	mqsixLog(">> mqsixGetAttributeName:  Index=%d\n", index);
	// Ensure that the attribute is within range
	if (index >= pAttributes->iAttributeCount)
	{
		mqsixLog("--   Index too large, highest: %d\n", pAttributes->iAttributeCount);
		return(1);
	}
	mqsixStringCopy(pAttributes->attributeNames[index], buffer, bufsize);
	mqsixLog("--   Returning name: %S\n", pAttributes->attributeNames[index]);
	return(0);
}

///////////////////////////////////////////////////////////
// mqsixGetAttribute
int mqsixGetAttribute(
	MQSIX_ATTRIBUTES	*pAttributes,
	CciChar				*attrName,
	CciChar				*buffer,
	int					bufsize)
{
	int i;
	mqsixLog(">> mqsixGetAttribute: %S\n", attrName);
	for (i=0; i<pAttributes->iAttributeCount; i++)
	{
		if (mqsixStringCompare(attrName, pAttributes->attributeNames[i]) == 0)
		{
			if (pAttributes->attributeValues[i] == NULL)
			{
				mqsixLog("--   Name found but no value\n");
				return(1);
			}
			mqsixStringCopy(pAttributes->attributeValues[i], buffer, bufsize);
			mqsixLog("--   Returning: \"%S\"\n", buffer);
			return(0);
		}
	}
	mqsixLog("--   No matching name found\n");
	return(1);
}

///////////////////////////////////////////////////////////
// mqsixSetAttribute
int mqsixSetAttribute(
	MQSIX_ATTRIBUTES	*pAttributes,
	CciChar				*attrName,
	CciChar				*attrValue)
{
	int	i;
	CciChar *temp;

	mqsixLog(">> mqsixSetAttribute: %S=\"%S\"\n", attrName, attrValue);
	for (i=0; i<pAttributes->iAttributeCount; i++)
	{
		if (mqsixStringCompare(attrName, pAttributes->attributeNames[i]) == 0)
		{
			temp = mqsixStringDuplicate(attrValue);
			if (pAttributes->attributeValues[i] != NULL)
			{
				free(pAttributes->attributeValues[i]);
			}
			pAttributes->attributeValues[i] = temp;
			return(0);
		}
	}
	return(1);
}

///////////////////////////////////////////////////////////
// mqsixLog
void mqsixLog(char *szFormat, ...)
{
#ifndef _DEBUG
	return;
#else
	static FILE	*logFile = NULL;
	struct tm	*now_tm;
	time_t		iNow;
	va_list		list;

	va_start(list, szFormat);

	if (logFile == NULL)
	{
		logFile = fopen(LOGFILE, "a");
		if (logFile == NULL)
		{
			return;
		}
	}

	iNow = time(NULL);
	now_tm = localtime(&iNow);
	fprintf(logFile, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d ",
		now_tm->tm_mon,
		now_tm->tm_mday,
		now_tm->tm_year+1900,
		now_tm->tm_hour,
		now_tm->tm_min,
		now_tm->tm_sec);
	vfprintf(logFile, szFormat, list);
	fflush(logFile);
	va_end(list);
#endif
}

///////////////////////////////////////////////////////////
// mqsixElementCharValue
CciSize mqsixElementCharValue(CciElement *pElement, CciChar **buffer)
{
	int	iRc;
	CciChar	oneChar;
	CciChar	*pRetBuffer = NULL;
	CciSize	iSize;
	iSize = cniElementCharacterValue(&iRc, pElement, &oneChar, sizeof(oneChar));

	if (iRc == CCI_BUFFER_TOO_SMALL || iRc == 0)
	{
		pRetBuffer = (CciChar *)malloc(iSize * sizeof(CciChar));
		iSize = cniElementCharacterValue(&iRc, pElement, pRetBuffer, iSize);
		//pRetBuffer[iSize] = 0;
		if (iRc != 0)
		{
			free(pRetBuffer);
			pRetBuffer = NULL;
			iSize = 0;
		}
	}
	*buffer = pRetBuffer;
	return(iSize);
}

///////////////////////////////////////////////////////////
// mqsixElementByteArrayValue
CciSize mqsixElementByteArrayValue(CciElement *pElement, struct CciByteArray *buffer)
{
	int	iRc;
	struct CciByteArray oneByteArray;
	char	oneByte;
	CciChar	*pRetBuffer = NULL;
	CciSize	iSize;

	oneByteArray.pointer = &oneByte;
	oneByteArray.size = 1;
	iSize = cniElementByteArrayValue(&iRc, pElement, &oneByteArray);

	if (iRc == CCI_BUFFER_TOO_SMALL || iRc == 0)
	{
		buffer->pointer = (void *)malloc(iSize);
		buffer->size = iSize;
		iSize = cniElementByteArrayValue(&iRc, pElement, buffer);
		//pRetBuffer[iSize] = 0;
		if (iRc != 0)
		{
			free(buffer->pointer);
			buffer->pointer = NULL;
			buffer->size = 0;
		}
	}
	return(iSize);
}

///////////////////////////////////////////////////////////
// mqsixElementToString
char *mqsixElementToString(CciElement *pElement)
{
	CciElementType	iElementType;
	CciValueType	iElementValueType;
	int				iRc;
	static char		text[1024];
	char			strValue[512];
	CciChar			name[256];
	int				i;
	char			twoHexChar[3];
	
	CciSize			iSize;
	CciBool			valueBool;
	CciInt			valueInt;
	CciReal			valueReal;
	CciChar			*valueChar;
	struct CciTime	valueTime;
	struct CciTime	valueGmtTime;
	struct CciDate	valueDate;
	struct CciTimestamp valueTimestamp;
	struct CciTimestamp valueGmtTimestamp;
	struct CciByteArray valueByteArray;



	iElementType = cniElementType(&iRc, pElement) & CCI_ELEMENT_TYPE_MASK_GENERIC;

	switch(iElementType)
	{
	case CCI_ELEMENT_TYPE_UNKNOWN:
		strcpy(text, "<** Unknown **>");
		break;
	case CCI_ELEMENT_TYPE_VALUE:
		strcpy(text, "= ");
		break;
	case CCI_ELEMENT_TYPE_NAME:
		cniElementName(&iRc, pElement, name, sizeof(name));
		sprintf(text, "<%S>", name);
		break;
	case CCI_ELEMENT_TYPE_NAME_VALUE:
		cniElementName(&iRc, pElement, name, sizeof(name));
		sprintf(text, "<%S> = ", name);
		break;
	}

	switch(iElementType)
	{
	case CCI_ELEMENT_TYPE_UNKNOWN:
	case CCI_ELEMENT_TYPE_NAME:
		break;
	case CCI_ELEMENT_TYPE_VALUE:
	case CCI_ELEMENT_TYPE_NAME_VALUE:
		iElementValueType = cniElementValueType(&iRc, pElement);
		switch(iElementValueType)
		{
		case CCI_VALUE_TYPE_UNKNOWN:
			strcat(text, "<** Unknown **>");
			break;
		case CCI_VALUE_TYPE_BOOL:
			valueBool = cniElementBooleanValue(&iRc, pElement);
			if (valueBool == 0)
			{
				strcat(text, "false");
			}
			else
			{
				strcat(text, "true");
			}
			break;
		case CCI_VALUE_TYPE_INTEGER:
			valueInt = cniElementIntegerValue(&iRc, pElement);
			//_i64toa(valueInt, strValue, 10);
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_REAL:
			valueReal = cniElementRealValue(&iRc, pElement);
			sprintf(strValue, "%f");
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_DECIMAL:
			// TBD
			break;
		case CCI_VALUE_TYPE_CHAR:
			// TBD
			iSize = mqsixElementCharValue(pElement, &valueChar);
			if (iSize == 0)
			{
				sprintf(strValue, "(Null)");
			}
			else
			{
				sprintf(strValue, "\"%S\"", valueChar);
				free(valueChar);
			}
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_TIME:
			valueTime = cniElementTimeValue(&iRc, pElement);
			sprintf(strValue, "TIME '%.2d:%.2d:%2.2f'",
				valueTime.hour, valueTime.minute, valueTime.second);
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_GMTTIME:
			valueGmtTime = cniElementGmtTimeValue(&iRc, pElement);
			sprintf(strValue, "GMTTIME '%.2d:%.2d:%2.2f'",
				valueGmtTime.hour, valueGmtTime.minute, valueGmtTime.second);
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_DATE:
			valueDate = cniElementDateValue(&iRc, pElement);
			sprintf(strValue, "DATE '%.4d-%.2d-%.2d'",
				valueDate.year, valueDate.month, valueDate.day);
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_TIMESTAMP:
			valueTimestamp = cniElementTimestampValue(&iRc, pElement);
			sprintf(strValue, "TIMESTAMP '%.4d-%.2d-%.2d %.2d:%.2d:%2.2f'",
				valueTimestamp.year, valueTimestamp.month, valueTimestamp.day,
				valueTimestamp.hour, valueTimestamp.minute, valueTimestamp.second);
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_GMTTIMESTAMP:
			valueGmtTimestamp = cniElementGmtTimestampValue(&iRc, pElement);
			sprintf(strValue, "GMTTIMESTAMP '%.4d-%.2d-%.2d %.2d:%.2d:%2.2f'",
				valueGmtTimestamp.year, valueGmtTimestamp.month, valueGmtTimestamp.day,
				valueGmtTimestamp.hour, valueGmtTimestamp.minute, valueGmtTimestamp.second);
			strcat(text, strValue);
			break;
		case CCI_VALUE_TYPE_BYTEARRAY:
			iSize = mqsixElementByteArrayValue(pElement, &valueByteArray);
			if (iSize == 0)
			{
				strcat(text, "(Null)");
			}
			else
			{
				strcat(text, "x'");
				for (i=0; i<valueByteArray.size; i++)
				{
					sprintf(twoHexChar, "%.2x", ((unsigned char *)valueByteArray.pointer)[i]);
					strcat(text, twoHexChar);
				}
				strcat(text, "'");
				free(valueByteArray.pointer);
			}
			break;
		case CCI_VALUE_TYPE_BITARRAY:
			break;
		}
		break;
	}
	return(text);
}

///////////////////////////////////////////////////////////
// mqsixWalkTree
void mqsixWalkTree(CciElement *pElement)
{
	CciElement	*child;
	int			iRc;

	child = cniFirstChild(&iRc, pElement);
	while(child != NULL)
	{
		mqsixWalkTree(child);
		child = cniNextSibling(&iRc, child);
	}
	return;
}

///////////////////////////////////////////////////////////
// mqsixLogTreeWithIndent
static void mqsixLogTreeWithIndent(CciElement *pElement, int iIndent)
{
	char		indentString[80];
	char		*text;
	int			iRc;
	int			i;
	CciElement	*child;

	text = mqsixElementToString(pElement);
	strcpy(indentString, "");
	for (i=0; i<iIndent; i++)
	{
		strcat(indentString, "  ");
	}
	mqsixLog("%s%s\n", indentString, text);

	child = cniFirstChild(&iRc, pElement);
	while(child != NULL)
	{
		mqsixLogTreeWithIndent(child, iIndent+1);
		child = cniNextSibling(&iRc, child);
	}
}

///////////////////////////////////////////////////////////
// mqsixLogTree
void mqsixLogTree(CciElement *pElement)
{
	mqsixLogTreeWithIndent(pElement, 0);
	return;
}

///////////////////////////////////////////////////////////
// mqsixLogInfo
void mqsixLogInfo(CciNode *pNode)
{
	int	iRc;
	CNI_BROKER_INFO_ST	brokerInfo;

	brokerInfo.brokerName.buffer = (CciChar *)malloc(sizeof(CciChar) * 100);
	brokerInfo.brokerName.bufferLength = 100;
	brokerInfo.executionGroupName.buffer = (CciChar *)malloc(sizeof(CciChar) * 100);
	brokerInfo.executionGroupName.bufferLength = 100;
	brokerInfo.messageFlowName.buffer = (CciChar *)malloc(sizeof(CciChar) * 100);
	brokerInfo.messageFlowName.bufferLength = 100;
	brokerInfo.queueManagerName.buffer = (CciChar *)malloc(sizeof(CciChar) * 100);
	brokerInfo.queueManagerName.bufferLength = 100;
	cniGetBrokerInfo(&iRc, pNode, &brokerInfo);
	mqsixLog("Version Id: %x\nBroker Name: %S\nExecution Group: %S\nMessage Flow: %S\nQueue Manager: %S\n",
		brokerInfo.versionId,
		brokerInfo.brokerName.buffer,
		brokerInfo.executionGroupName.buffer,
		brokerInfo.messageFlowName.buffer,
		brokerInfo.queueManagerName.buffer);
	free(brokerInfo.brokerName.buffer);
	free(brokerInfo.executionGroupName.buffer);
	free(brokerInfo.messageFlowName.buffer);
	free(brokerInfo.queueManagerName.buffer);
}
